/**
 * \file flowmon-export-ipfix.h
 * \author Petr Velan <velan@mail.muni.cz>
 * \date 2012
 * \brief Header for the code of the Flowmon IPFIX Export Plugin.
 *
 * \copyright
 * Copyright (C) 2012 Masaryk University, Institute of Computer Science
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Masaryk University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id: flowmon-export-ipfix.h 858 2012-12-12 10:34:02Z 255519 $
 */

#include <time.h>
#include <netdb.h>

#ifdef DEBUG
#define MSG(format, ...) fprintf(stderr, format, ## __VA_ARGS__)
#else
#define MSG(format, ...) {}
#endif

#define DEFAULT_TEMPLATE_FILE "/etc/flowmon/ipfix-template-file.txt"
#define FIRST_TEMPLATE_ID 258
#define IPFIX_VERISON 10
#define TEMPLATE_SET_ID 2
#define PACKET_DATA_SIZE 1458 /* ethernet 14, ip 20, udp 8 */
/* Template buffer cannot be filled more than to packet data size */
#define TEMPLATE_BUFFER_SIZE (PACKET_DATA_SIZE - IPFIX_HEADER_SIZE)
#define IPFIX_HEADER_SIZE 16
#define IPFIX_SET_HEADER_SIZE 4
#define RECONNECT_TIMEOUT 60

/**
 * \brief Structure to hold one entry of template-file
 *
 * GETTER_NAME ENTERPRISE_NUMBER ELEMENT_ID LENGTH
 */
typedef struct {
	char *getterName; /**< FlowMon getter name */
	uint16_t enterpriseNumber; /**< Enterprise Number */
	uint16_t elementID; /**< Information Element ID */
	int32_t length; /**< Element export length. -1 for variable*/
} template_file_record_t;


/**
 * \brief Structure to hold template record
 */
typedef struct template_t {
	uint16_t id; /**< Template ID */
	uint8_t templateRecord[1024]; /**< Buffer for template record, 516B is big
									enough for 64 enterprise elements */
	uint16_t templateSize; /**< Size of template record buffer */
	uint8_t buffer[TEMPLATE_BUFFER_SIZE]; /**< Buffer with data for template */
	uint16_t bufferSize; /**< Size of data buffer */
	uint16_t recordCount; /**< Number of records in buffer */
	uint16_t fieldCount; /**< Number of elements in template */
	uint64_t mask; /**< bitmask of getter array for this template */
	flow_record_getter_t *templateGetters; /**< Getters for this template */
	template_file_record_t **templateFileRecords; /**< Template elements from file */
	uint8_t exported; /**< 1 indicates that the template was exported to collector*/
	time_t exportTime; /**< Time when the template was last exported */
	uint64_t exportPacket; /**< Number of packet when the template was last exported */
	struct template_t *next;
} template_t;


/**
 * \brief Structure that holds plugin internal data between function calls
 */
typedef struct {
	/* General */
	flow_record_getter_t *getters; /**< List of used getters */
	uint32_t sequenceNum; /**< Number of exported flows */
	uint64_t counter; /**< Number of exported flows -- to be removed*/
	uint64_t exportedPackets; /**< Number of exported packets */
	int socket; /**< Socket used to send data */
	struct addrinfo *addrinfo; /**< Info about the connection used by sendto */

	/* Parameters */
	char *host; /**< Hostname */
	char *templateFile; /**< Template file */
	char *port; /**< Port */
	int protocol; /**< Protocol */
	int ip; /**< IP protocol version (AF_INET, ...) */
	int flags; /**< getaddrinfo flags */
	uint32_t reconnectTimeout; /**< Timeout between connection retries */
	time_t lastReconnect; /**< Time in seconds of last connection retry */
	uint32_t odid; /**< Observation Domain ID */
	uint32_t templateRefreshTime; /**< UDP template refresh time interval */
	uint32_t templateRefreshPackets; /**< UDP template refresh packet interval */

	/* Templates */
	template_file_record_t *templateFileRecords; /**< File template records  */
	template_t *templates; /**< Templates in use by plugin */
	uint16_t templatesDataSize; /**< Total data size stored in templates */
} plugin_private_t;


/**
 * \brief Structure of ipfix packet used by send functions
 */
typedef struct {
	char *data; /**< Buffer for data */
	uint16_t length; /**< Length of data */
	uint16_t flows; /**< Number of flow records in the packet */
} ipfix_packet_t;


/**
 * \brief IPFIX header structure
 *
 *   0                   1                   2                   3
 *   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |       Version Number          |            Length             |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                           Export Time                         |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                       Sequence Number                         |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                    Observation Domain ID                      |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 */
typedef struct ipfix_header {
	/**
	 * Version of Flow Record format exported in this message. The value of this
	 * field is 0x000a for the current version, incrementing by one the version
	 * used in the NetFlow services export version 9.
	 */
	uint16_t version;

	/**
	 * Total length of the IPFIX Message, measured in octets, including Message
	 * Header and Set(s).
	 */
	uint16_t length;

	/**
	 * Time, in seconds, since 0000 UTC Jan 1, 1970, at which the IPFIX Message
	 * Header leaves the Exporter.
	 */
	uint32_t exportTime;

	/**
	 * Incremental sequence counter modulo 2^32 of all IPFIX Data Records sent
	 * on this PR-SCTP stream from the current Observation Domain by the
	 * Exporting Process. Check the specific meaning of this field in the
	 * subsections of Section 10 when UDP or TCP is selected as the transport
	 * protocol. This value SHOULD be used by the Collecting Process to
	 * identify whether any IPFIX Data Records have been missed. Template and
	 * Options Template Records do not increase the Sequence Number.
	 */
	uint32_t sequenceNumber;

	/**
	 * A 32-bit identifier of the Observation Domain that is locally unique to
	 * the Exporting Process. The Exporting Process uses the Observation Domain
	 * ID to uniquely identify to the Collecting Process the Observation Domain
	 * that metered the Flows. It is RECOMMENDED that this identifier also be
	 * unique per IPFIX Device. Collecting Processes SHOULD use the Transport
	 * Session and the Observation Domain ID field to separate different export
	 * streams originating from the same Exporting Process. The Observation
	 * Domain ID SHOULD be 0 when no specific Observation Domain ID is relevant
	 * for the entire IPFIX Message, for example, when exporting the Exporting
	 * Process Statistics, or in case of a hierarchy of Collectors when
	 * aggregated Data Records are exported.
	 */
	uint32_t observationDomainId;
} ipfix_header_t;

/**
 * \brief Common IPFIX Set (header) structure
 *
 *   0                   1                   2                   3
 *   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |          Set ID               |          Length               |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 */
typedef struct ipfix_template_set_header {
	/**
	 * Set ID value identifies the Set.  A value of 2 is reserved for the
	 * Template Set. A value of 3 is reserved for the Option Template Set. All
	 * other values from 4 to 255 are reserved for future use. Values above 255
	 * are used for Data Sets. The Set ID values of 0 and 1 are not used for
	 * historical reasons [<a href="http://tools.ietf.org/html/rfc3954">RFC3954</a>].
	 */
	uint16_t id;

	/**
	 * Total length of the Set, in octets, including the Set Header, all
	 * records, and the optional padding.  Because an individual Set MAY contain
	 * multiple records, the Length value MUST be used to determine the position
	 * of the next Set.
	 */
	uint16_t length;

} ipfix_template_set_header_t;

/**
 *  Plugin description variable
 */
static plugin_desc_t plugin_desc = {
	"ipfixx",
	"Plugin that exports data using IPFIX protocol\n"
	"Parameters are (delimited by ','):\n"
	" host=hostname                (default = localhost)\n"
	" port=port                    (default = 4739)\n"
	" ip=ip                        IP version used to connect to collector (4,6,64 or default).\n"
	"                              64 uses IPv6 and then IPv4 mapped addresses, default takes priorities from the system.\n"
	" protocol=proto               Protocol used to connect to collector - udp(default),tcp,sctp.\n"
	" reconnect-timeout=seconds    Timeout in seconds between reconnect attempts.\n"
	" odid=id                      Observation Domain ID (default = 0).\n"
	" template-file=filename       File with template parameters. (default = /etc/flowmon/ipfix-template-file.txt).\n"
	" template-refresh-time=time   Send templates to collector each 'time' seconds. UDP only (default = 600).\n"
	" template-refresh-packets=num Send templates to collector each 'num' packets. UDP only.\n",
	0,
	0
};
